import os
from concurrent.futures import Future, ThreadPoolExecutor
from typing import TYPE_CHECKING, Any, Generator, Optional

import llguidance  # type: ignore[import-untyped]
import numpy as np
from numpy.typing import NDArray

from ._schema import GenData, LegacyEngineCallResponse, LLInterpreterResponse

if TYPE_CHECKING:
    from .models._engine import Tokenizer


class TokenParserException(Exception):
    pass


class InvalidTokenException(TokenParserException):
    def __init__(self, token: int, valid_tokens: list[int]):
        self.token = token
        self.valid_tokens = valid_tokens
        super().__init__(f"Invalid token {token}, expected one of {valid_tokens}")


class TokenParser:
    def __init__(
        self,
        grammar: str,
        tokenizer: "Tokenizer",
        enable_backtrack: bool = True,
        enable_ff_tokens: bool = True,
    ):
        self.tokenizer = tokenizer
        self.ll_tokenizer = tokenizer._ll_tokenizer
        self.ll_interpreter = llguidance.LLInterpreter(
            self.ll_tokenizer,
            grammar,
            enable_backtrack,
            enable_ff_tokens,
            log_level=int(os.environ.get("LLGUIDANCE_LOG_LEVEL", "1")),
        )
        self._threadpool = ThreadPoolExecutor(max_workers=1)
        self._generator = self._parse()
        self._done = False
        self._has_pending_stop = False

    def is_accepting(self) -> bool:
        return self.ll_interpreter.is_accepting()

    def done(self) -> bool:
        return self._done

    def advance(
        self, token_id: Optional[int]
    ) -> tuple[
        int,
        list[int],
        Future[tuple[Optional[bytes], LLInterpreterResponse]],
    ]:
        if self.done():
            raise TokenParserException("Cannot advance on a done parser")

        return self._generator.send(token_id)

    def has_pending_stop(self) -> bool:
        return self._has_pending_stop

    def process_prompt(
        self, prompt_tokens: list[int], ensure_bos_token: bool = True
    ) -> tuple[
        list[int],  # prefix_tokens
        int,  # backtrack
        list[int],  # ff_tokens
        # mask: Optional[bytes] (None if stop), ll_response: LLInterpreterResponse
        Future[tuple[Optional[bytes], LLInterpreterResponse]],
    ]:
        new_prompt_tokens = self.ll_interpreter.process_prompt(prompt_tokens)
        if (
            ensure_bos_token
            and self.tokenizer.bos_token_id is not None
            and new_prompt_tokens[:1] != [self.tokenizer.bos_token_id]
        ):
            assert prompt_tokens[:1] != [self.tokenizer.bos_token_id]
            prefix_tokens = [self.tokenizer.bos_token_id]
        else:
            prefix_tokens = []

        _backtrack, _ff_tokens, mask_fut = self._generator.send(None)
        assert _backtrack == 0
        assert _ff_tokens == []

        # Apply the prefix tokens before computing the backtrack
        prompt_tokens = prefix_tokens + prompt_tokens
        new_prompt_tokens = prefix_tokens + new_prompt_tokens

        # Compute backtrack and ff_tokens s.t.
        # new_prompt_tokens == (
        #       prompt_tokens[:-backtrack] if backtrack > 0
        #       else prompt_tokens
        #   ) + ff_tokens
        backtrack = len(prompt_tokens)
        for old, new in zip(prompt_tokens, new_prompt_tokens):
            if old != new:
                break
            backtrack -= 1
        common_len = len(prompt_tokens) - backtrack
        ff_tokens = new_prompt_tokens[common_len:]

        return prefix_tokens, backtrack, ff_tokens, mask_fut

    def compute_mask(self) -> tuple[Optional[bytes], LLInterpreterResponse]:
        mask, ll_response_string = self.ll_interpreter.compute_mask()
        ll_response = LLInterpreterResponse.model_validate_json(ll_response_string)
        return mask, ll_response

    def _parse(
        self,
    ) -> Generator[
        tuple[
            int,  # backtrack
            list[int],  # ff_tokens
            # mask: Optional[bytes] (None if stop), ll_response: LLInterpreterResponse
            Future[tuple[Optional[bytes], LLInterpreterResponse]],
        ],
        Optional[int],
        None,
    ]:
        backtrack = 0
        ff_tokens: list[int] = []
        token_id: Optional[int] = None
        while True:
            # Note: need to call/set has_pending_stop before spinning up the compute mask
            # future as the two methods cannot be called concurrently
            self._has_pending_stop = self.ll_interpreter.has_pending_stop()
            compute_mask_future = self._threadpool.submit(self.compute_mask)

            # Send caller the mask and response; wait for token
            token_id = yield (backtrack, ff_tokens, compute_mask_future)

            # Upstairs should have already waited on this future
            mask, r = compute_mask_future.result()

            if r.stop:
                # This is the only case in which the mask is None
                assert mask is None
                # If we're done, our caller should NOT send us a token
                if token_id is not None:
                    raise TokenParserException(f"Expected None, got token {token_id}")
                self._done = True
                break

            assert mask is not None
            assert r.temperature is not None

            if token_id is None:
                raise TokenParserException("Expected token, got None")

            if not mask[token_id]:
                # Note: we could punt this probem to ll_interpreter.post_process,
                # but it's a bit clearer to handle it here
                raise InvalidTokenException(
                    token=token_id,
                    valid_tokens=[i for i in range(len(mask)) if mask[i]],
                )

            backtrack, ff_tokens = self.ll_interpreter.commit_token(token_id)

    def cleanup(self):
        # Rather than having our caller send us None at the end, we'll handle that internally
        # so we can (1) verify that the generator actually stops and (2) check the stop reason
        # and raise if needed
        if not self.done():
            try:
                self._generator.send(None)
            except StopIteration:
                pass
        if not self.done():
            raise TokenParserException("Tried to cleanup but parser is not done")
        stop_reason = self.ll_interpreter.stop_reason()
        if stop_reason not in {"NoExtension", "EndOfSentence"}:
            # Will raise if there is some "bad" stop reason (like hit token limit) OR we're NOT stopped.
            # TODO: raise specific exceptions for reasons such as MaxTokensTotal
            raise TokenParserException(f"Unexpected stop reason: {stop_reason}")


class ByteParserException(Exception):
    def __init__(self, *args, **kwargs):
        self.current_byte = kwargs.pop("current_byte", None)
        self.allowed_bytes = kwargs.pop("allowed_bytes", None)
        self.consumed_bytes = kwargs.pop("consumed_bytes", None)
        super().__init__(*args, **kwargs)


class ByteParser:
    def __init__(
        self,
        grammar: str,
    ):
        # TODO: figure out this circular import
        from .models._byte_tokenizer import ByteTokenizer

        self.tokenizer = ByteTokenizer()
        self.token_parser = TokenParser(grammar, self.tokenizer)
        self.bytes = b""
        self.gen_data: Optional[GenData] = None
        self.pos = 0
        self._variables: dict[str, Any] = {}
        self._variables_log_probs: dict[str, Any] = {}
        # Prime the parser
        self._advance(None)

    def matched(self) -> bool:
        if self.pos < len(self.bytes):
            return False
        return self.token_parser.is_accepting()

    def valid_next_bytes(self) -> set[bytes]:
        if self.pos < len(self.bytes):
            return {self.bytes[self.pos : self.pos + 1]}
        if self.gen_data is None:
            return set()
        return {bytes([t]) for t in self.gen_data.valid_next_tokens if t != self.tokenizer.eos_token_id}

    def next_byte_mask(self) -> NDArray[np.uint8]:
        mask = np.zeros(256, dtype=np.uint8)
        for t in self.valid_next_bytes():
            mask[t[0]] = 1
        return mask

    def _advance(self, token_id: Optional[int]) -> None:
        if self.gen_data is not None:
            tokens = self.gen_data.tokens
        else:
            tokens = []
        if token_id is None:
            assert tokens == []
            prefix_tokens, backtrack, ff_tokens, compute_mask_future = self.token_parser.process_prompt(tokens)
            tokens += prefix_tokens
        else:
            backtrack, ff_tokens, compute_mask_future = self.token_parser.advance(token_id)
        if backtrack:
            tokens = tokens[:-backtrack]
        tokens += ff_tokens
        mask, ll_response = compute_mask_future.result()
        if ll_response.stop:
            assert mask is None
            self.token_parser.cleanup()
            self.gen_data = None
        else:
            assert mask is not None
            assert ll_response.temperature is not None
            self.gen_data = GenData(
                tokens=tokens,
                mask=mask,
                temperature=ll_response.temperature,
            )
        response = ll_response.progress.to_engine_call_response()
        self._update_capture(response)
        self.bytes += response.new_bytes

    def consume_bytes(self, bts: bytes) -> None:
        if not bts:
            return

        b = bts[0]
        # If the current position is less than the length of the bytes, then we are in fast_forward mode
        # and we need to make sure that the byte we are consuming is the same as the byte at the current
        # position
        if self.pos < len(self.bytes):
            if b != self.bytes[self.pos]:
                next_byte = self.bytes[self.pos : self.pos + 1]
                raise ByteParserException(
                    f"Expected byte {next_byte!r} (fast_forward), got {bytes([b])!r}",
                    current_byte=bytes([b]),
                    allowed_bytes={next_byte},
                    consumed_bytes=self.bytes[: self.pos],
                )
            # Byte was good, move to the next byte
            self.pos += 1
            self.consume_bytes(bts[1:])
        else:
            # If we are here, then we are either in generation mode or we are done.
            if self.gen_data is None:
                # TODO: may run into trouble here if we need to backtrack
                assert self.token_parser.done()
                assert not self.valid_next_bytes()
                raise ByteParserException(
                    f"Expected end of input, got {bytes([b])!r}",
                    current_byte=bytes([b]),
                    allowed_bytes=set(),
                    consumed_bytes=self.bytes[: self.pos],
                )
            # We're in generation mode. Assure that the byte is one of the valid next bytes
            if b not in self.gen_data.valid_next_tokens:
                valid_next_bytes = self.valid_next_bytes()
                raise ByteParserException(
                    f"Expected one of the following bytes: {valid_next_bytes!r}, got {bytes([b])!r}",
                    current_byte=bytes([b]),
                    allowed_bytes=valid_next_bytes,
                    consumed_bytes=self.bytes[: self.pos],
                )
            # Byte was good, have ll_parser consume it so we can advance further
            self._advance(b)

            # Run consume_bytes to advance ll_parser and consume the next byte
            self.consume_bytes(bts)

    def force_done(self):
        if not self.matched():
            raise ByteParserException("Hit end of input before reaching a valid state")
        if self.token_parser.done():
            return

        self._advance(self.tokenizer.eos_token_id)
        if not self.token_parser.done() or not self.matched():
            raise ByteParserException("Hit end of input before reaching a valid state")

    def get_captures(self):
        return self._variables, self._variables_log_probs

    def _update_capture(self, response: LegacyEngineCallResponse):
        # Stolen from model. TODO: refactor to share code
        for k in response.capture_groups:
            v = response.capture_groups[k]

            # see if we are in a list_append mode
            if isinstance(v, list):
                for i, inner_v in enumerate(v):
                    # convert to a string if possible
                    # TODO: will need to not just always do this once we support images etc.
                    try:
                        inner_v = inner_v.decode("utf8") if isinstance(inner_v, bytes) else inner_v
                    except UnicodeDecodeError:
                        pass

                    if k not in self._variables or not isinstance(self._variables[k], list):
                        self._variables[k] = []
                        self._variables_log_probs[k] = []
                    self._variables[k].append(inner_v)
                    self._variables_log_probs[k].append(response.capture_group_log_probs[k][i])

            # ...or standard assignment mode
            else:
                # convert to a string if possible
                # TODO: will need to not just always do this once we support images etc.
                try:
                    v = v.decode("utf8") if isinstance(v, bytes) else v
                except UnicodeDecodeError:
                    pass
                self._variables[k] = v
                self._variables_log_probs[k] = response.capture_group_log_probs[k]



from typing import Optional

from guidance._schema import SamplingParams

from ._base import Model
from ._openai_base import (
    BaseOpenAIInterpreter,
    OpenAIAudioMixin,
    OpenAIClientWrapper,
    OpenAIImageMixin,
    OpenAIJSONMixin,
    OpenAIRegexMixin,
    OpenAIRuleMixin,
)


class OpenAIInterpreter(OpenAIRuleMixin, OpenAIJSONMixin, OpenAIRegexMixin, BaseOpenAIInterpreter):
    def __init__(
        self,
        model: str,
        api_key: Optional[str] = None,
        **kwargs,
    ):
        try:
            import openai
        except ImportError as ie:
            raise Exception(
                "Please install the openai package version >= 1 using `pip install openai -U` in order to use guidance.models.OpenAI!"
            ) from ie

        client = openai.OpenAI(api_key=api_key, **kwargs)
        super().__init__(model=model, client=OpenAIClientWrapper(client))


class OpenAI(Model):
    def __init__(
        self,
        model: str,
        sampling_params: Optional[SamplingParams] = None,
        echo: bool = True,
        *,
        api_key: Optional[str] = None,
        **kwargs,
    ):
        """Build a new OpenAI model object that represents a model in a given state.

        Parameters
        ----------
        model : str
            The name of the OpenAI model to use (e.g. gpt-4o-mini).
        echo : bool
            If true the final result of creating this model state will be displayed (as HTML in a notebook).
        api_key : None or str
            The OpenAI API key to use for remote requests, passed directly to the `openai.OpenAI` constructor.

        **kwargs :
            All extra keyword arguments are passed directly to the `openai.OpenAI` constructor. Commonly used argument
            names include `base_url` and `organization`
        """

        if "audio-preview" in model:
            interpreter_cls = type("OpenAIAudioInterpreter", (OpenAIAudioMixin, OpenAIInterpreter), {})
        elif model.startswith("gpt-4o") or model.startswith("o1"):
            interpreter_cls = type("OpenAIImageInterpreter", (OpenAIImageMixin, OpenAIInterpreter), {})
        else:
            interpreter_cls = OpenAIInterpreter

        super().__init__(
            interpreter=interpreter_cls(model, api_key=api_key, **kwargs),
            sampling_params=SamplingParams() if sampling_params is None else sampling_params,
            echo=echo,
        )



import re
from typing import Callable, Iterable, Literal, Optional, Union

from .._ast import RuleNode, SubstringNode


def chunk_on_word(text: str) -> list[str]:
    return re.findall(r"(\s+|\w+|[^\s\w]+)", text)


def substring(
    target_string: str,
    *,
    chunk: Union[Literal["word", "character"], Callable[[str], Iterable[str]]] = "word",
    name: Optional[str] = None,
) -> RuleNode:
    chunks: Iterable[str]
    if chunk == "word":
        chunks = chunk_on_word(target_string)
    elif chunk == "character":
        chunks = tuple(target_string)
    elif callable(chunk):
        chunks = chunk(target_string)
        if "".join(chunks) != target_string:
            raise ValueError(
                "chunk_on function must return a sequence of strings that can be joined to form the target string"
            )
    else:
        raise ValueError(f"Invalid `chunk` value: {chunk!r}. Expected 'word', 'character', or a function.")

    return RuleNode(
        name=name or "substring",
        value=SubstringNode(tuple(chunks)),
        capture=name,
    )



"""Debug utilities for the guidance widget."""

import logging
from typing import Optional

from .registry import get_renderer
from .visual._renderer import AutoRenderer, JupyterWidgetRenderer

logger = logging.getLogger(__name__)


def enable_widget_debug() -> None:
    """Enable debug mode in the guidance widget.

    This will start capturing all messages sent to the widget for later inspection.
    Call dump_widget_debug() to dump the captured data to a string.

    Example:
        >>> import guidance
        >>> guidance.enable_widget_debug()
        >>> # ... run your guidance code ...
        >>> guidance.dump_widget_debug()
    """
    renderer = get_renderer()
    if isinstance(renderer, JupyterWidgetRenderer):
        renderer.enable_debug()
    elif isinstance(renderer, AutoRenderer):
        # Check if the auto renderer has a JupyterWidgetRenderer inside
        if hasattr(renderer, "_renderer") and isinstance(renderer._renderer, JupyterWidgetRenderer):
            renderer._renderer.enable_debug()
        else:
            logger.warning(f"Auto renderer contains {type(renderer._renderer)} instead of JupyterWidgetRenderer")
    else:
        logger.warning(f"Debug mode only available with Jupyter widget renderer, got {type(renderer)}")


def dump_widget_debug() -> Optional[str]:
    """Get captured debug data as a JSON string.

    Returns the captured widget messages and state as a JSON string,
    or None if no debug data is available.

    Example:
        >>> import guidance
        >>> guidance.enable_widget_debug()
        >>> # ... run your guidance code ...
        >>> debug_data = guidance.dump_widget_debug()
        >>> print(debug_data)  # inspect in notebook
    """
    renderer = get_renderer()
    if isinstance(renderer, JupyterWidgetRenderer):
        return renderer.get_debug_data()
    elif isinstance(renderer, AutoRenderer):
        # Check if the auto renderer has a JupyterWidgetRenderer inside
        if hasattr(renderer, "_renderer") and isinstance(renderer._renderer, JupyterWidgetRenderer):
            return renderer._renderer.get_debug_data()
        else:
            logger.warning(f"Auto renderer contains {type(renderer._renderer)} instead of JupyterWidgetRenderer")
            return None
    else:
        logger.warning(f"Debug dump only available with Jupyter widget renderer, got {type(renderer)}")
        return None


def clear_widget_debug() -> None:
    """Clear captured widget debug messages.

    Useful for starting fresh between different test runs.

    Example:
        >>> import guidance
        >>> guidance.enable_widget_debug()
        >>> # ... run some guidance code ...
        >>> guidance.clear_widget_debug()  # clear before next test
        >>> # ... run different guidance code ...
        >>> debug_data = guidance.dump_widget_debug()
    """
    renderer = get_renderer()
    if isinstance(renderer, JupyterWidgetRenderer):
        renderer.clear_debug_data()
    elif isinstance(renderer, AutoRenderer):
        # Check if the auto renderer has a JupyterWidgetRenderer inside
        if hasattr(renderer, "_renderer") and isinstance(renderer._renderer, JupyterWidgetRenderer):
            renderer._renderer.clear_debug_data()
        else:
            logger.warning(f"Auto renderer contains {type(renderer._renderer)} instead of JupyterWidgetRenderer")
    else:
        logger.warning(f"Clear debug only available with Jupyter widget renderer, got {type(renderer)}")


def widget_debug_info() -> None:
    """Print debug information about the current widget renderer setup.

    This function displays Python-side version info and requests version info
    from the JavaScript widget if available.
    """
    import platform
    import sys

    from .visual._environment import Environment

    renderer = get_renderer()
    env = Environment()

    print(f"Renderer type: {type(renderer)}")
    print(f"Environment is_notebook(): {env.is_notebook()}")
    print(f"Environment is_terminal(): {env.is_terminal()}")
    print(f"Python version: {sys.version.split()[0]}")
    print(f"Platform: {platform.system()} {platform.release()}")

    # Check stitch installation (Python side)
    try:
        import stitch

        version = getattr(stitch, "__version__", "unknown")
        print(f"Stitch (Python): {version}")
    except ImportError:
        print("Stitch (Python): Not installed")

    # Check guidance version
    try:
        import guidance

        guidance_version = getattr(guidance, "__version__", "unknown")
        print(f"Guidance version: {guidance_version}")
    except ImportError:
        guidance_version = "unknown"
        print("Guidance version: unknown")

    # Try to get widget version info if we have a JupyterWidgetRenderer
    widget_renderer = None
    if isinstance(renderer, JupyterWidgetRenderer):
        widget_renderer = renderer
        print("Widget debug mode should work!")
    elif isinstance(renderer, AutoRenderer):
        print(f"AutoRenderer inner type: {type(renderer._renderer)}")
        if hasattr(renderer, "_renderer") and isinstance(renderer._renderer, JupyterWidgetRenderer):
            widget_renderer = renderer._renderer
            print("Widget debug mode should work!")
        else:
            print("Widget debug mode not available - inner renderer is not JupyterWidgetRenderer")
    else:
        print("Widget debug mode not available - not using widget renderer")



from typing import Any

from ..._schema import TokenUsage
from .._base import State


class EngineState(State):
    def __init__(self) -> None:
        # Initialize with zero token usage rather than default None
        # since engine can fast-forward tokens
        super().__init__(token_usage=TokenUsage(ff_tokens=0))
        self.prompt: str = ""
        self.images: list[Any] = []
        self.audio: list[Any] = []
        self.videos: list[Any] = []

    def __str__(self) -> str:
        return self.prompt



import logging
from typing import Optional, Sequence

import numpy as np

from .._schema import GenTokenExtra, SamplingParams
from ._base import Model
from ._engine import Engine, EngineInterpreter, LogitsOutput, Tokenizer
from ._engine._tokenizer import TokenizerWrappable

logger = logging.getLogger(__name__)


class MockTokenizer(Tokenizer):
    def __init__(self, tokens: Sequence[bytes], special_token_ids: Optional[list[int]] = None):
        self.tokens = tokens
        self.byte_trie = ByteTrie(self.tokens, np.arange(len(self.tokens)))

        ll_tokenizer = TokenizerWrappable(
            eos_token_id=0,
            bos_token_id=0,
            tokens=tokens,
            special_token_ids=[0],
            # ENCODE MUST BE OVERRIDDEN
            encode_callable=self.encode,
        ).as_ll_tokenizer()

        super().__init__(
            ll_tokenizer=ll_tokenizer,
            chat_template=None,
            bos_token_id=0,
        )

    def encode(self, byte_string: bytes, *, parse_special: bool = True) -> list[int]:
        """Simple greedy tokenizer
        TODO: could be a method on ByteTrie if we want to reuse it
        """
        if not parse_special:
            raise ValueError("parse_special=False is not supported in MockTokenizer")
        pos = 0
        tokens = []
        while pos < len(byte_string):
            current_node = self.byte_trie
            last_match = None
            match_pos = pos

            while match_pos < len(byte_string) and current_node.has_child(byte_string[match_pos : match_pos + 1]):
                current_node = current_node.child(byte_string[match_pos : match_pos + 1])
                if current_node.value >= 0:
                    last_match = (current_node.value, match_pos + 1)
                match_pos += 1

            if last_match is not None:
                tokens.append(last_match[0])
                pos = last_match[1]
            else:
                raise ValueError(f"Could not find a match for byte {byte_string[pos]} at position {pos}")

        return tokens

    def recode(self, tokens: Sequence[int]) -> list[int]:
        # Make a no-op for now
        return list(tokens)


class MockEngine(Engine):
    def __init__(self, tokenizer, byte_patterns, force):
        super().__init__(tokenizer)

        self._valid_mask = np.zeros(len(tokenizer.tokens))
        for i, t in enumerate(tokenizer.tokens):
            try:
                t.decode("utf8")
                self._valid_mask[i] = 1.0
            except:
                pass
        self.force = force
        self.called_temperatures = []

        # allow a single byte pattern to be passed
        if isinstance(byte_patterns, (bytes, str)):
            byte_patterns = [byte_patterns]

        # allow for strings to be passed
        for i, pattern in enumerate(byte_patterns):
            if isinstance(pattern, str):
                byte_patterns[i] = pattern.encode("utf8")

        self.byte_patterns = byte_patterns

        # seed the random number generator
        self._rand_generator = np.random.default_rng(seed=42)

    def get_next_token_with_top_k(
        self,
        logits: Optional[np.ndarray],
        logits_lat_ms: Optional[float],
        token_ids: list[int],
        mask: Optional[bytes],
        temperature: float,
        k: int = 1,
        force_return_unmasked_probs: bool = False,
        sampling_params: Optional[SamplingParams] = None,
    ) -> GenTokenExtra:
        self.called_temperatures.append(temperature)
        return super().get_next_token_with_top_k(
            logits, logits_lat_ms, token_ids, mask, temperature, k, force_return_unmasked_probs, sampling_params
        )

    def get_logits(self, token_ids: list[int], include_all_uncached_tokens: bool = False) -> LogitsOutput:
        """Pretends to compute the logits for the given token state."""
        if len(token_ids) == 0:
            raise ValueError("token_ids must not be empty")

        # build the byte strings
        byte_string = b"".join(self.tokenizer.tokens[i] for i in token_ids)

        # if we are forcing the bytes patterns then don't allow other tokens
        if self.force:
            logits = np.ones(len(self.tokenizer.tokens)) * -np.inf

        # otherwise we randomly generate valid unicode bytes
        else:
            logits = self._rand_generator.standard_normal(len(self.tokenizer.tokens)) * self._valid_mask

        # if we have a pattern that matches then force the next token
        bias = 100.0
        if self.byte_patterns is not None:
            for p in self.byte_patterns:
                if p.startswith(byte_string) and len(p) > len(byte_string):
                    for i in self._get_next_tokens(p[len(byte_string) :]):
                        logits[i] += bias
                        bias /= 2  # if we have multiple matches then they apply with decreasing bias

        return {
            "logits": logits.reshape(1, -1),
            "n_tokens": len(token_ids),
            # TODO: add caching support and report this number accurately?
            # This might help mock some tests that make assertions about caching.
            "n_cached": 0,
        }

    def _get_next_tokens(self, byte_string):
        special_tokens = [
            (self.tokenizer.bos_token_id, self.tokenizer.bos_token),
            (self.tokenizer.eos_token_id, self.tokenizer.eos_token),
        ]
        for i, t in special_tokens:
            # if the byte string starts with a special token then make sure we don't yield any other tokens
            if byte_string.startswith(t):
                yield i
                return
        for i, t in enumerate(self.tokenizer.tokens):
            if byte_string.startswith(t):
                yield i


class Mock(Model):
    def __init__(
        self,
        byte_patterns=[],
        sampling_params: Optional[SamplingParams] = None,
        echo=False,
        force=False,
        **kwargs,
    ):
        """Build a new Mock model object that represents a model in a given state."""

        # Our tokens are all bytes and all lowercase letter pairs
        all_lc_pairs = [bytes([i, j]) for i in range(ord("a"), ord("z")) for j in range(ord("a"), ord("z"))]
        all_bytes = [bytes([i]) for i in range(256)]
        tokens = [b"<s>"] + all_lc_pairs + all_bytes

        tokenizer = MockTokenizer(tokens, special_token_ids=[0])
        engine = MockEngine(tokenizer, byte_patterns, force)

        super().__init__(
            interpreter=EngineInterpreter(engine),
            echo=echo,
            sampling_params=SamplingParams() if sampling_params is None else sampling_params,
        )


class ByteTrie:
    """A python implementation mirroring the C++ ByteTrie class."""

    def __init__(self, byte_strings=None, values=None, parent=None):
        self._parent = parent
        self.match_version = -1
        self.match = False
        self.partial_match = False
        self.prob = 0
        self.value = -1
        self.children = {}

        if byte_strings is not None:
            if values is None:
                for s in byte_strings:
                    self.insert(s, 0)
            else:
                for i, s in enumerate(byte_strings):
                    self.insert(s, values[i])

    def keys(self):
        return self.children.keys()

    def has_child(self, byte):
        return byte in self.children

    def child(self, byte):
        return self.children[byte]

    def parent(self):
        return self._parent

    def size(self):
        return len(self.children)

    def __len__(self):
        return self.size()

    def insert(self, s, value, pos=0):
        if len(s) <= pos:
            if self.value < 0:
                self.value = value
        else:
            first_byte = s[pos : pos + 1]
            if first_byte not in self.children:
                self.children[first_byte] = ByteTrie(parent=self)
            self.children[first_byte].insert(s, value, pos + 1)

    def compute_probs(self, probs):
        self.prob = 0.0

        if self.value != -1:
            self.prob += probs[self.value]

        if self.children:
            for k in self.children:
                child = self.children[k]
                child.compute_probs(probs)
                self.prob += child.prob



import asyncio
import logging
import time
from asyncio import CancelledError
from enum import Enum
from typing import Any, Sequence

import psutil

from .._schema import TokenUsage
from .._topics import METRICS_TOPIC
from ..visual import MetricMessage

MISSING_VALUE = 0

logger = logging.getLogger(__name__)


class PeriodicMetricsGenerator:
    def __init__(self, monitor: "Monitor", sleep_sec=0.5):
        self._monitor = monitor
        self._sleep_sec = sleep_sec
        self._task = None
        self._cancelled = False
        self._is_paused = False

    def start(self):
        from ..registry import get_bg_async

        bg = get_bg_async()
        self._task = bg.run_async_coroutine(bg.async_task(self._emit())).result()

    def stop(self):
        if self._task is not None:
            self._cancelled = True
            self._task.cancel()

    def pause(self):
        """
        Pauses the model by setting the internal _is_paused flag to True.

        This method can be used to temporarily halt the model's operations.
        """
        self._is_paused = True

    def resume(self):
        """
        Resume the model's operation by setting the paused state to False.

        This method changes the internal state of the model to indicate that it is no longer paused.
        """
        self._is_paused = False

    async def _emit(self):
        import asyncio

        from ..registry import get_exchange

        while not self._cancelled:
            try:
                await asyncio.sleep(self._sleep_sec)

                cpu_percent = self._monitor.get_metric(MonitoringMetric.CPU_USAGE)
                used_ram = self._monitor.get_metric(MonitoringMetric.MEM_USAGE)
                gpu_percent = self._monitor.get_metric(MonitoringMetric.GPU_USAGE)
                gpu_used_vram = self._monitor.get_metric(MonitoringMetric.GPU_USED_MEM)

                if gpu_percent:
                    gpu_percent = max(gpu_percent)
                else:
                    gpu_percent = MISSING_VALUE

                if gpu_used_vram:
                    gpu_used_vram = max(gpu_used_vram)
                else:
                    gpu_used_vram = MISSING_VALUE

                if not cpu_percent:
                    cpu_percent = MISSING_VALUE

                if not used_ram:
                    used_ram = MISSING_VALUE

                if not self._is_paused:
                    exchange = get_exchange()
                    exchange.publish(MetricMessage(name="cpu", value=cpu_percent), topic=METRICS_TOPIC)
                    exchange.publish(MetricMessage(name="ram", value=used_ram), topic=METRICS_TOPIC)
                    exchange.publish(MetricMessage(name="gpu", value=gpu_percent), topic=METRICS_TOPIC)
                    exchange.publish(MetricMessage(name="vram", value=gpu_used_vram), topic=METRICS_TOPIC)
            except CancelledError:
                logger.debug("METRICGEN:canceling")
                break
            except Exception as e:  # noqa BLE001
                logger.debug(f"METRICGEN: {repr(e)}")
                break

        logger.debug("METRICGEN:exiting")


class MonitoringMetric(str, Enum):
    CPU_USAGE = "cpu_usage"
    MEM_USAGE = "mem_usage"
    GPU_USAGE = "gpu_usage"
    GPU_USED_MEM = "gpu_used_mem"
    GPU_TOTAL_MEM = "gpu_total_mem"


class Monitor:
    """Monitoring service to collect necessary metrics for visualization"""

    def __init__(self, interval_ms: int = 1000, **kwargs):
        self.max_size = kwargs.get("max_size", 100)
        self.stop_flag = False
        self.task = None
        self.interval_ms = interval_ms

        # Initialize metrics storage
        self.metrics_dict = {
            MonitoringMetric.CPU_USAGE: [],
            MonitoringMetric.MEM_USAGE: [],
            MonitoringMetric.GPU_USAGE: [],
            MonitoringMetric.GPU_USED_MEM: [],
            MonitoringMetric.GPU_TOTAL_MEM: [],
        }

    async def _monitor_fn(self):
        import traceback

        to_collect_gpu_stats = False
        has_gpustat = False
        try:
            import gpustat

            has_gpustat = True
        except ImportError:
            logger.warning("gpustat is not installed, run `pip install gpustat` to collect GPU stats.")

        if has_gpustat:
            try:
                gpu_stats = gpustat.GPUStatCollection.new_query()
                if len(gpu_stats) > 0:
                    to_collect_gpu_stats = True
            except Exception as e:  # noqa BLE001
                logger.warning(f"Non-Nvidia GPU monitoring is not supported in this version. {e}")

        while not self.stop_flag:
            try:
                t0 = time.time()

                cpu_per_core_percent = psutil.cpu_percent(interval=None, percpu=True)
                average_cpu_percent = sum(cpu_per_core_percent) / len(cpu_per_core_percent)
                average_cpu_utilization = average_cpu_percent / 100.0
                memory_usage = psutil.virtual_memory()
                memory_usage_gb = memory_usage.used / (1024**3)

                self.metrics_dict[MonitoringMetric.CPU_USAGE].append(average_cpu_utilization)
                self.metrics_dict[MonitoringMetric.MEM_USAGE].append(memory_usage_gb)
                if to_collect_gpu_stats:
                    gpu_stats = gpustat.GPUStatCollection.new_query()

                    usage = [gpu.utilization / 100.0 for gpu in gpu_stats.gpus]
                    mem_usage = [gpu.memory_used for gpu in gpu_stats.gpus]
                    mem_total = [gpu.memory_total for gpu in gpu_stats.gpus]

                    self.metrics_dict[MonitoringMetric.GPU_USAGE].append(usage)
                    self.metrics_dict[MonitoringMetric.GPU_USED_MEM].append(mem_usage)
                    self.metrics_dict[MonitoringMetric.GPU_TOTAL_MEM].append(mem_total)

                # Trim lists to max_size
                for metrics in self.metrics_dict.values():
                    if len(metrics) > self.max_size:
                        metrics = metrics[-self.max_size :]

                lat = time.time() - t0
                sleep_time = self.interval_ms / 1000.0 - lat
                if sleep_time > 0:
                    await asyncio.sleep(sleep_time)

            except Exception as e:  # noqa BLE001
                logger.error(f"Caught {e}")
                logger.error(f"MONITOR:{traceback.format_exc()}")
                await asyncio.sleep(1)  # Wait a bit before retrying on error

    def start(self):
        from ..registry import get_bg_async

        bg = get_bg_async()
        self.stop_flag = False
        self.task = bg.run_async_coroutine(bg.async_task(self._monitor_fn())).result()
        logger.debug("MONITOR:start")

    def stop(self):
        self.stop_flag = True
        if self.task:
            self.task.cancel()
            self.task = None
        for metrics in self.metrics_dict.values():
            metrics.clear()
        logger.debug("MONITOR:stop")

    def reset(self):
        self.stop()
        self.start()
        logger.debug("MONITOR:reset")

    def get_metrics(
        self,
        metrics: Sequence[MonitoringMetric] = (),
    ) -> dict[MonitoringMetric, Any]:
        if not metrics:
            metrics = MonitoringMetric.__members__.values()
        result = {}
        for metric in metrics:
            if metric in MonitoringMetric.__members__.values():
                result[metric] = self.metrics_dict[metric][-1] if len(self.metrics_dict[metric]) > 0 else None
            else:
                raise ValueError(f"Unknown monitoring metric: {metric}")
        return result

    def get_metric(self, metric: MonitoringMetric) -> Any:
        return self.get_metrics([metric])[metric]


def emit_usage(usage: TokenUsage) -> None:
    from ..registry import get_exchange

    exchange = get_exchange()

    exchange.publish(
        MetricMessage(
            name="token reduction",
            value=(usage.token_savings or 0) * 100,  # display as percentage
        ),
        topic=METRICS_TOPIC,
    )

    exchange.publish(MetricMessage(name="consumed", value=usage.forward_passes), topic=METRICS_TOPIC)

    exchange.publish(MetricMessage(name="avg latency", value=usage.avg_latency_ms), topic=METRICS_TOPIC)



from .._ast import GrammarNode, RuleNode
from .._grammar import regex, subgrammar

__all__ = ["as_regular_grammar", "lexeme", "regex", "subgrammar"]


def as_regular_grammar(node: GrammarNode, lexeme=False):
    # TODO: Remove this assertion-only check?
    if isinstance(node, RuleNode):
        rule = node
    else:
        rule = RuleNode("dummy", node)
    assert rule.is_allowed_in_lark_terminal
    return node


def lexeme(body_regex: str, json_string: bool = False):
    if json_string:
        raise NotImplementedError("JSON strings are not supported")
    return regex(body_regex)



from . import experimental
from ._azureai import create_azure_aifoundry_model, create_azure_openai_model
from ._base import Model
from ._llama_cpp import LlamaCpp
from ._mock import Mock
from ._openai import OpenAI
from ._transformers import Transformers

__all__ = [
    "LlamaCpp",
    "Mock",
    "Model",
    "OpenAI",
    "Transformers",
    "create_azure_aifoundry_model",
    "create_azure_openai_model",
    "experimental",
]



"""Metrics that arise from both language models and its execution environment."""

from ._metrics import Monitor, PeriodicMetricsGenerator, emit_usage

__all__ = [
    "Monitor",
    "PeriodicMetricsGenerator",
    "emit_usage",
]



import logging
from typing import Optional, Union

from .._grammar import capture, quote_regex, select, with_temperature
from .._grammar import gen as grammar_gen
from .._grammar import regex as regex_node
from .._guidance import guidance
from ._block import block
from ._tool import Tool

logger = logging.getLogger(__name__)


def gen(
    name=None,
    *,
    max_tokens=None,
    list_append=False,
    regex=None,
    tools=None,
    hide_tool_call=False,
    stop: Union[str, list[str], None] = None,
    stop_regex: Union[str, list[str], None] = None,
    suffix: Optional[str] = None,
    n=1,
    temperature=None,
    top_p=1.0,
    save_stop_text: Union[bool, str] = False,
):
    """Generate a set of tokens until a given stop criteria has been met.

    This function is a useful utility that can allow you to specify most grammars used by typical
    LM generation programs. It also has the added ability to interleave generation with tool calls.

        >>> lm += gen("my_generation", max_tokens=10)
        >>> print(lm["my_generation"])
        some text from the LLM

    Parameters
    ----------

        name : str or None
            If this is not None then the the results of the generation will be saved as a variable on
            the Model object (so you can access the result as `lm["var_name"]`).

        max_tokens : int
            The maximum number of generation tokens we should use. Note that this limit is not exact when
            regular expression pattern constraints are present, but guidance does attempt to end the generation
            as soon as possible while keeping the regex constraints satisfied.

        list_append : bool
            If this is True then the results saved to `lm[name]` will not be written directly but rather appended
            to a list (if no list with the current name is present one will be created). This is useful for
            building lists inside python loops.

        regex : str or None
            This is a regular expression that will be used to constrain the generation. The model is only allowed
            to generate tokens that match this regular expression. Note that for variable length expressions the
            model is free to continue the expression after a complete match, but generation will terminate as soon
            as the model generates anything that does not match the pattern (this ending behavior may change a bit we
            update guidance to maintain the grammar parsing state between calls).

        stop : str or list or None
            The stop string (or list of strings) we should use for terminating this generation segment.

        stop_regex : str or list or None
            The stop regular expression (or list of regular expressions) we should use for terminating this generation segment.

        save_stop_text : bool or str
            If True then this saves the captured stop text or regex into a variable of the name `str(name) + "_stop_text"`. If
            a string is given then the captured stop text is saved under that name.

        temperature : float
            The temperature to use during this generation call. Note that when parsing ambiguous grammars that include
            multiple conflicting temperatures (for example from multiple possible `gen` calls inside a `select`) the highest
            temperature of all options is used by the model (since we only want to run the model once, not once for every
            possible parse path).

        top_p : float
            TODO! Will control the models top_p generation parameter, but has been yet been implemented beyond top_p=1.0.

        n : int
            TODO! Will control the number of parallel generation calls made during gen.

        tools : Tool or list or None
            A list of `guidance.Tool` or Python functions (which will be converted to `guidance.Tool`).
            When using `tools` you must specify `max_tokens` and cannot use `regex`.

        hide_tool_call : bool
            Controls if we should hide the text generated by the model to trigger a tool call. You may want to hide the tool
            call from the model's context if you plan to change its format after the call is made.
    """
    if tools and not max_tokens:
        raise ValueError("Must specify `max_tokens` when using `tools`")
    if tools and regex:
        raise ValueError("Cannot use `regex` with `tools`")

    assert n == 1, "We still need to add support for n>1! Consider putting your gen call in a loop for now."
    assert top_p == 1, "Please use `model.with_sampling_params` to set top_p."

    logger.debug(f'start gen(name="{name}")')

    if stop is not None and stop_regex is not None:
        raise ValueError("Cannot use both stop and stop_regex")
    if isinstance(stop, list):
        stop_regex = [quote_regex(s) for s in stop]
        stop = None
    if isinstance(stop_regex, list):
        stop_regex = "|".join([r for r in stop_regex])

    if save_stop_text is False:
        save_stop_name = None
    elif save_stop_text is True:
        # TODO: "None_stop_text" -- is that really what we want?
        save_stop_name = str(name) + "_stop_text"
    else:
        save_stop_name = save_stop_text

    if tools is not None:
        tools = [Tool(callable=x) if not isinstance(x, Tool) else x for x in tools]

        @guidance(stateless=False, dedent=False)
        def tool_gen(lm):
            options = [
                grammar_gen(
                    regex=regex,
                    stop_regex=stop_regex,
                    stop=stop,
                    stop_capture=save_stop_name,
                    temperature=temperature,
                    max_tokens=max_tokens,
                )
            ]
            for i, tool in enumerate(tools):
                # Infer a regex that will match the start of a tool call
                tool_call_prefix = tool.call_grammar.forced_prefix()
                if len(tool_call_prefix) < 4:
                    # TODO: alternatively check that the prefix contains the name (case insensitive) of the tool?
                    # anything shorter is probably far too ambiguous
                    raise ValueError(f"Could not infer unambiguous tool call prefix for tool {tool.name}")
                options.append(
                    grammar_gen(
                        regex=regex,
                        stop=tool_call_prefix,
                        temperature=temperature,
                        max_tokens=max_tokens,
                        name=f"tool{i}",
                    ),
                )

            grm = with_temperature(select(options), temperature)
            initial_token_count = lm._get_usage().output_tokens
            tagged_name = "__LIST_APPEND:" + name if list_append and name is not None else name
            with block(tagged_name):
                while lm._get_usage().output_tokens <= max_tokens + initial_token_count:
                    lm += grm
                    tool_called = False
                    for i in range(len(tools)):
                        tool_i = f"tool{i}"
                        if tool_i in lm:
                            tool_called = True
                            if hide_tool_call:
                                temp_lm = lm + tools[i].call_grammar
                                with block("tool_call"):
                                    temp_lm += tools[i].tool_call()
                                lm += temp_lm["tool_call"]
                            else:
                                lm += tools[i].call_grammar + tools[i].tool_call()
                            lm = lm.remove(tool_i)
                    if not tool_called:
                        if suffix is not None:
                            lm += suffix
                        break
            return lm

        return tool_gen()

    return grammar_gen(
        regex=regex,
        stop_regex=stop_regex,
        stop=stop,
        suffix=suffix,
        stop_capture=save_stop_name,
        name=name,
        list_append=list_append,
        temperature=temperature,
        max_tokens=max_tokens,
    )


@guidance
def call_tool(lm, tool):
    lm += tool.call_grammar
    lm += tool.tool_call()
    return lm


@guidance(stateless=True)
def regex(lm, pattern, *, name=None):
    node = regex_node(pattern)
    if name:
        node = capture(node, name)
    return lm + node



import base64
import importlib.resources
import pathlib
import re
import typing

from .._ast import ImageBlob, ImageUrl
from .._guidance import guidance
from .._utils import bytes_from
from ..trace._trace import ImageOutput


@guidance
def image(lm, src: typing.Union[str, pathlib.Path, bytes], allow_local: bool = True):
    if isinstance(src, str) and re.match(r"^(?!file://)[^:/]+://", src):
        lm += ImageUrl(url=src)
    else:
        bytes_data = bytes_from(src, allow_local=allow_local)
        base64_bytes = base64.b64encode(bytes_data)
        lm += ImageBlob(data=base64_bytes)
    return lm


@guidance
def gen_image(lm):
    # TODO(nopdive): Mock for testing. Remove all of this code later.
    with importlib.resources.files("guidance").joinpath("resources/sample_image.png").open("rb") as f:
        bytes_data = f.read()
    base64_bytes = base64.b64encode(bytes_data)
    lm += ImageOutput(value=base64_bytes, is_input=False)
    return lm



from contextlib import contextmanager
from typing import Optional, Union

from .._ast import ASTNode, Function
from .._guidance import _in_stateless_context
from ..models._base._model import _active_blocks


class Block:
    def __init__(
        self, name: Optional[str], opener: Union[str, Function, ASTNode], closer: Union[str, Function, ASTNode]
    ):
        self.name = name
        self.opener = opener
        self.closer = closer


@contextmanager
def block(name=None, opener=None, closer=None):
    if _in_stateless_context.get():
        raise RuntimeError("Cannot use roles or other blocks when stateless=True")
    current_blocks = _active_blocks.get()
    new_block = Block(name=name, opener=opener, closer=closer)
    token = _active_blocks.set(current_blocks + (new_block,))
    try:
        yield
    finally:
        _active_blocks.reset(token)



"""Registry module that contains singletons."""

from ._registry import get_bg_async, get_exchange, get_monitor, get_renderer, get_trace_handler, set_renderer

__all__ = [
    "get_bg_async",
    "get_exchange",
    "get_monitor",
    "get_renderer",
    "get_trace_handler",
    "set_renderer",
]



from typing import TYPE_CHECKING, Any, ContextManager, Iterator, Optional

from guidance._schema import SamplingParams

from ..._ast import GrammarNode, JsonNode, RegexNode, RuleNode
from ...trace import OutputAttr, TextOutput
from .._base import Model
from .._openai_base import BaseOpenAIClientWrapper, BaseOpenAIInterpreter

if TYPE_CHECKING:
    from openai.types.chat import ChatCompletionChunk
import contextlib


class LiteLLMOpenAIClientWrapper(BaseOpenAIClientWrapper):
    def __init__(self, router):
        self.router = router

    @contextlib.contextmanager
    def _wrapped_completion(
        self,
        model: str,
        messages: list[dict[str, Any]],
        logprobs: bool,
        **kwargs,
    ) -> Iterator["ChatCompletionChunk"]:
        """Wrapped completion call within a context manager."""
        kwargs["stream"] = True  # Ensure we are streaming here
        stream_wrapper = self.router.completion(
            model=model,
            messages=messages,
            logprobs=logprobs,
            **kwargs,
        )

        try:
            yield stream_wrapper
        finally:
            # stream_wrapper.completion_stream is the underlying stream, e.g. openai.Stream
            if hasattr(stream_wrapper.completion_stream, "close"):
                # Close the stream if it has a close method
                stream_wrapper.completion_stream.close()

    def streaming_chat_completions(
        self,
        model: str,
        messages: list[dict[str, Any]],
        logprobs: bool,
        **kwargs,
    ) -> ContextManager[Iterator["ChatCompletionChunk"]]:
        """Streaming chat completions."""
        return self._wrapped_completion(
            model=model,
            messages=messages,
            logprobs=logprobs,
            **kwargs,
        )  # type: ignore[return-value]


class LiteLLMInterpreter(BaseOpenAIInterpreter):
    SUPPORTED_ENDPOINT_TYPES = ["openai", "azure_ai", "azure", "gemini", "anthropic", "xai", "hosted_vllm"]

    def __init__(self, model_description: dict, **kwargs):
        try:
            import litellm
        except ImportError as ie:
            raise Exception(
                "Please install the litellm package version >= 1.71.0 using `pip install litellm -U` in order to use guidance.models.LiteLLM!"
            ) from ie

        self.ep_type = self._check_model(model_description)
        # set default model to the first one in the list
        self.model = model_description["model_name"]
        self.router = litellm.Router(model_list=[model_description])
        self.client = LiteLLMOpenAIClientWrapper(self.router)

        # Disable log_probs for any remote endpoints by default.
        # Otherwise, generation will fail for some endpoints.
        self.logprobs = False

        super().__init__(model=self.model, client=self.client, **kwargs)

    def _check_model(self, model_desc: dict) -> str:
        """Check if the model description is valid."""
        for ep_type in self.SUPPORTED_ENDPOINT_TYPES:
            if model_desc["litellm_params"]["model"].startswith(ep_type):
                return ep_type

        raise Exception(
            "Cannot parse endpoint type. "
            "Please use this 'model' format in 'litellm_params': endpoint_type/model_name (e.g., openai/gpt-3.5-turbo). "
            "Supported endpoints are: " + ", ".join(self.SUPPORTED_ENDPOINT_TYPES)
        )

    def rule(self, node: RuleNode, **kwargs) -> Iterator[OutputAttr]:
        kwargs = self._process_kwargs(**kwargs)

        # Disable this check for now as all the supported endpoints have 'stop' support.
        # if node.stop and self.ep_type not in ["hosted_vllm", "azure", "azure_ai", "gemini", "openai", "xai", "anthropic"]:
        #     raise ValueError(f"stop condition not yet supported for {self.ep_type} endpoint")
        if node.suffix:
            raise ValueError(f"suffix not yet supported for {self.ep_type} endpoint")
        if node.stop_capture:
            raise ValueError(f"stop_capture not yet supported for {self.ep_type} endpoint")

        kwargs = kwargs.copy()
        if node.temperature:
            kwargs["temperature"] = node.temperature
        if node.max_tokens:
            kwargs["max_tokens"] = node.max_tokens
        if node.stop:
            if self.ep_type in ["xai"] and isinstance(node.stop.regex, str):
                kwargs["stop"] = [node.stop.regex]
            else:
                kwargs["stop"] = node.stop.regex

        chunks = self.run(node.value, **kwargs)
        if node.capture:
            buffered_text = ""
            for chunk in chunks:
                # TODO: this isinstance check is pretty darn fragile.
                # ~there must be a better way~
                if isinstance(chunk, TextOutput):
                    buffered_text += chunk.value
                yield chunk
            yield self.state.apply_capture(
                name=node.capture,
                value=buffered_text,
                log_prob=1,  # TODO
                is_append=node.list_append,
            )
        else:
            yield from chunks

    def regex(self, node: RegexNode, **kwargs) -> Iterator[OutputAttr]:
        kwargs = self._process_kwargs(**kwargs)
        if node.regex is not None and self.ep_type not in ["hosted_vllm"]:
            raise ValueError(f"Regex is not yet supported for ep {self.ep_type}")

        if self.ep_type == "hosted_vllm":
            return self._regex_vllm(node, **kwargs)

        # We're in unconstrained mode now.
        return self._run(**kwargs)

    def _regex_vllm(self, node: RegexNode, **kwargs):
        """Run the regex node using vLLM."""
        if "extra_body" not in kwargs:
            kwargs["extra_body"] = {}

        kwargs["extra_body"].update(dict(guided_decoding_backend="guidance", guided_regex=node.regex))

        buffer: str = ""
        for attr in self._run(**kwargs):
            if isinstance(attr, TextOutput):
                buffer += attr.value
            yield attr

    def json(self, node: JsonNode, **kwargs) -> Iterator[OutputAttr]:
        kwargs = self._process_kwargs(**kwargs)
        if (
            self.ep_type in ["openai"]
            and (self.model in ["gpt-3.5-turbo"] or self.model.startswith("gpt-4-"))
            and node.schema is not None
        ):
            raise ValueError(f"json_schema is not supported for ep {self.ep_type}/{self.model}")

        if self.ep_type in ["azure_ai"]:
            raise ValueError(f"json_object/json_schema are not supported for ep {self.ep_type}")

        if node.schema is None:
            response_format = {"type": "json_object"}
        else:
            # set additionalProperties to False but allow it to be overridden
            node.schema["additionalProperties"] = node.schema.get("additionalProperties", False)
            response_format = {
                "type": "json_schema",
                "json_schema": {
                    "name": "json_schema",
                    "schema": node.schema,
                    "strict": True,
                },
            }

        return self._run(
            response_format=response_format,
            **kwargs,
        )

    def grammar(self, node: GrammarNode, **kwargs) -> Iterator[OutputAttr]:
        kwargs = self._process_kwargs(**kwargs)
        if self.ep_type == "hosted_vllm":
            return self._grammar_vllm(node, **kwargs)

        raise ValueError(f"Grammar is not yet supported for ep {self.ep_type}")

    def _grammar_vllm(self, node: GrammarNode, **kwargs) -> Iterator[OutputAttr]:
        """Run the grammar node using vLLM."""
        if "extra_body" not in kwargs:
            kwargs["extra_body"] = {}

        kwargs["extra_body"].update(dict(guided_decoding_backend="guidance", guided_grammar=node.ll_grammar()))

        buffer: str = ""
        for attr in self._run(**kwargs):
            if isinstance(attr, TextOutput):
                buffer += attr.value
            yield attr
        matches = node.match(
            buffer,
            raise_exceptions=False,
            # Turn of max_tokens since we don't have access to the tokenizer
            enforce_max_tokens=False,
        )
        if matches is None:
            # TODO: should probably raise...
            # raise ValueError("vLLM failed to constrain the grammar")
            pass
        else:
            for name, value in matches.captures.items():
                log_probs = matches.log_probs[name]
                if isinstance(value, list):
                    assert isinstance(log_probs, list)
                    assert len(value) == len(log_probs)
                    for v, l in zip(value, log_probs):
                        yield self.state.apply_capture(name=name, value=v, log_prob=l, is_append=True)
                else:
                    yield self.state.apply_capture(name=name, value=value, log_prob=log_probs, is_append=False)

    def _process_kwargs(self, **kwargs):
        sampling_params = kwargs.pop("sampling_params", None)
        if sampling_params is None:
            return kwargs

        kwargs["top_p"] = sampling_params.pop("top_p", None)
        kwargs["top_k"] = sampling_params.pop("top_k", None)
        kwargs["min_p"] = sampling_params.pop("min_p", None)
        kwargs["repetition_penalty"] = sampling_params.pop("repetition_penalty", None)

        return kwargs


class LiteLLM(Model):
    def __init__(
        self, model_description: dict, sampling_params: Optional[SamplingParams] = None, echo: bool = True, **kwargs
    ):
        interpreter = LiteLLMInterpreter(model_description=model_description, **kwargs)
        super().__init__(
            interpreter=interpreter,
            sampling_params=SamplingParams() if sampling_params is None else sampling_params,
            echo=echo,
        )



import re
from base64 import b64decode, b64encode
from copy import deepcopy
from io import BytesIO
from typing import Iterator

from ..._ast import GrammarNode, ImageBlob, JoinNode, LiteralNode, RoleEnd, RoleStart, SpecialToken
from ..._schema import GenTokenExtra, TokenUsage
from ..._utils import to_utf8_or_bytes_string
from ...trace import Backtrack, ImageOutput, OutputAttr, Token, TokenOutput
from .._base import Interpreter
from ._engine import Engine, Tokenizer
from ._state import EngineState


class EngineInterpreter(Interpreter[EngineState]):
    def __init__(self, engine: Engine):
        super().__init__(state=EngineState())
        self.engine = engine
        self.chat_template = self.engine.get_chat_template()

    def __deepcopy__(self, memo):
        """Custom deepcopy to ensure engine is not copied."""
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == "engine":
                # Don't copy the engine
                setattr(result, k, v)
            else:
                setattr(result, k, deepcopy(v, memo))
        return result

    def get_role_start(self, role: str) -> str:
        if self.chat_template is None:
            raise ValueError("Cannot use roles without a chat template")
        return self.chat_template.get_role_start(role)

    def get_role_end(self, role: str) -> str:
        if self.chat_template is None:
            raise ValueError("Cannot use roles without a chat template")
        return self.chat_template.get_role_end(role)

    def role_start(self, node: RoleStart, **kwargs) -> Iterator[OutputAttr]:
        self.state.active_role = node.role
        text = self.get_role_start(node.role)
        # TODO: it's probably somewhat wasteful to trigger engine calls here,
        # so we can maybe add this as "pending text" to the state instead,
        # accumulating it until the next engine call..?
        yield from self.run(text_to_grammar(self.engine.tokenizer, text))

    def role_end(self, node: RoleEnd, **kwargs) -> Iterator[OutputAttr]:
        self.state.active_role = None
        text = self.get_role_end(node.role)
        # TODO: it's probably somewhat wasteful to trigger engine calls here,
        # so we can maybe add this as "pending text" to the state instead,
        # accumulating it until the next engine call..?
        yield from self.run(text_to_grammar(self.engine.tokenizer, text))

    def text(self, node: LiteralNode, **kwargs) -> Iterator[OutputAttr]:
        yield from self.grammar(node, **kwargs)

    def grammar(self, node: GrammarNode, **kwargs) -> Iterator[OutputAttr]:
        engine_gen = self.engine(
            state=self.state,
            grammar=node.ll_grammar(),
            ensure_bos_token=True,
            echo=False,
            sampling_params=kwargs.pop("sampling_params", None),
        )

        delayed_bytes = b""
        while True:
            try:
                chunk = next(engine_gen)
            except StopIteration as e:
                if not isinstance(e.value, TokenUsage):
                    raise e
                self.state.add_usage(e.value)
                break

            new_bytes = recode_special_tokens(self.engine.tokenizer, chunk.new_bytes)
            new_text, delayed_bytes = partial_decode(delayed_bytes + new_bytes)
            self.state.prompt += new_text

            if chunk.backtrack:
                yield Backtrack(
                    n_tokens=chunk.backtrack,
                    bytes=b64encode(chunk.backtrack_bytes),
                )

            for token in chunk.tokens:
                if isinstance(token, GenTokenExtra):
                    top_k = [
                        Token(
                            token=to_utf8_or_bytes_string(t.bytes),
                            bytes=b64encode(t.bytes),
                            prob=t.prob,
                            masked=t.is_masked,
                        )
                        for t in token.top_k
                    ]
                else:
                    top_k = None

                token_value = to_utf8_or_bytes_string(token.bytes)
                yield TokenOutput(
                    value=token_value,
                    token=Token(token=token_value, bytes=b64encode(token.bytes), prob=token.prob),
                    latency_ms=token.latency_ms,
                    is_input=token.is_input,
                    is_generated=token.is_generated,
                    is_force_forwarded=token.is_force_forwarded,
                    top_k=top_k,
                )
                if token.is_backtracked:
                    yield Backtrack(
                        n_tokens=1,
                        bytes=b64encode(token.bytes),
                    )

            for name in chunk.capture_groups.keys():
                values = chunk.capture_groups[name]
                log_probs = chunk.capture_group_log_probs[name]
                if isinstance(values, list):
                    assert isinstance(log_probs, list)
                    assert len(values) == len(log_probs)
                    for value, log_prob in zip(values, log_probs):
                        yield self.state.apply_capture(name, value.decode("utf-8"), log_prob=log_prob, is_append=True)
                else:
                    assert isinstance(log_probs, float)
                    yield self.state.apply_capture(name, values.decode("utf-8"), log_prob=log_probs, is_append=False)

        if delayed_bytes:
            raise RuntimeError("Shouldn't have any delayed bytes left...")


class Llama3VisionInterpreter(EngineInterpreter):
    def image_blob(self, node: ImageBlob, **kwargs) -> Iterator[OutputAttr]:
        try:
            import PIL.Image
        except ImportError as ie:
            raise Exception(
                "Please install the Pillow package `pip install Pillow` in order to use images with Llama3!"
            ) from ie

        image_bytes = b64decode(node.data)
        pil_image = PIL.Image.open(BytesIO(image_bytes))
        self.state.images.append(pil_image)
        self.state.prompt += "<|image|>"

        yield ImageOutput(value=node.data, is_input=True)


class Phi3VisionInterpreter(EngineInterpreter):
    def image_blob(self, node: ImageBlob, **kwargs) -> Iterator[OutputAttr]:
        try:
            import PIL.Image
        except ImportError as ie:
            raise Exception(
                "Please install the Pillow package `pip install Pillow` in order to use images with Llama3!"
            ) from ie

        image_bytes = b64decode(node.data)
        pil_image = PIL.Image.open(BytesIO(image_bytes))

        if pil_image in self.state.images:
            ix = self.state.images.index(pil_image) + 1
        else:
            self.state.images.append(pil_image)
            ix = len(self.state.images)
        self.state.prompt += f"<|image_{ix}|>"

        yield ImageOutput(value=node.data, is_input=True)


def partial_decode(data: bytes) -> tuple[str, bytes]:
    try:
        return (data.decode("utf-8"), b"")
    except UnicodeDecodeError as e:
        valid_part = data[: e.start].decode("utf-8")
        delayed_part = data[e.start :]
    return (valid_part, delayed_part)


LLG_SPECIAL_TOKEN_PAT = re.compile(rb"\xff\[([0-9]+)\]")


def recode_special_tokens(tokenizer: Tokenizer, data: bytes) -> bytes:
    """Recode a byte string with special tokens in llguidance format to their actual byte representation."""
    return LLG_SPECIAL_TOKEN_PAT.sub(lambda m: tokenizer.decode([int(m.group(1).decode("utf-8"))]), data)


def text_to_grammar(tokenizer: Tokenizer, text: str) -> GrammarNode:
    """
    Convert a text string into a GrammarNode that can be used in the grammar.
    This is useful for converting static text into a grammar node that can be processed by the engine.
    """
    grammar_bits: list[GrammarNode] = []
    delayed_bytes = b""
    for token_id in tokenizer.encode(text.encode("utf-8"), parse_special=True):
        if tokenizer.is_special_token(token_id):
            assert not delayed_bytes, "Should not have any delayed bytes when encountering a special token"
            grammar_bits.append(SpecialToken(id=token_id))
        else:
            new_bytes = tokenizer.decode([token_id])
            new_text, delayed_bytes = partial_decode(delayed_bytes + new_bytes)
            if new_text:
                grammar_bits.append(LiteralNode(new_text))
    assert not delayed_bytes, "Should not have any delayed bytes left after processing the text"
    if len(grammar_bits) == 1:
        return grammar_bits[0]
    return JoinNode(tuple(grammar_bits))



"""UI and other visual UX considerations.

Users should have few reasons to be accessing this module.
"""

from ._exchange import TopicExchange
from ._message import (
    ClientReadyAckMessage,
    ClientReadyMessage,
    ExecutionCompletedMessage,
    ExecutionStartedMessage,
    GuidanceMessage,
    MetricMessage,
    OutputRequestMessage,
    ResetDisplayMessage,
    TraceMessage,
    deserialize_message,
    serialize_message,
)
from ._renderer import AutoRenderer, JupyterWidgetRenderer, Renderer
from ._trace import display_trace_tree, trace_node_to_html, trace_node_to_str

__all__ = [
    "AutoRenderer",
    "ClientReadyAckMessage",
    "ClientReadyMessage",
    "ExecutionCompletedMessage",
    "ExecutionStartedMessage",
    "GuidanceMessage",
    "JupyterWidgetRenderer",
    "MetricMessage",
    "OutputRequestMessage",
    "Renderer",
    "ResetDisplayMessage",
    "TopicExchange",
    "TraceMessage",
    "deserialize_message",
    "display_trace_tree",
    "serialize_message",
    "trace_node_to_html",
    "trace_node_to_str",
]



"""Exchange topic constants for guidance message routing.

This module centralizes all topic constants used throughout the guidance
codebase for the TopicExchange message routing system.
"""

# Default topic for general message routing
DEFAULT_TOPIC = "/default"

# Metrics-related topics
METRICS_TOPIC = "/metrics"

# Trace topics
TRACE_TOPIC = "/trace"

# Visual topics
VISUAL_TOPIC = "/visual"

__all__ = [
    "DEFAULT_TOPIC",
    "METRICS_TOPIC",
    "TRACE_TOPIC",
    "VISUAL_TOPIC",
]



# TODO(nopdive): This module requires a memory review.

import logging
import time
from abc import ABC, abstractmethod
from typing import Generator, Optional, TypedDict

import numpy as np
from numpy.typing import NDArray

from ..._parser import TokenParser
from ..._schema import EngineResponse, GenToken, GenTokenExtra, SamplingParams, TokenUsage
from ..._utils import apply_min_p_filter, apply_repetition_penalty, apply_top_k_and_top_p_filter, log_init, softmax
from ._state import EngineState
from ._tokenizer import Tokenizer

logger = logging.getLogger(__name__)

_TEMPERATURE_EPSILON = 0.0001


class LogitsOutput(TypedDict):
    logits: NDArray
    n_tokens: int
    n_cached: int


class Engine(ABC):
    """The engine owns the inference computation and is used/created by the Model class.

    Engine objects represent the expensive parts of inference. While Model objects are cheap and do not
    need to know about the tokenizer or the model parameters, Engine objects know about both. Many
    Model objects can reference a single Engine object. Engine objects can also be hidden behind a
    Server so a single server can serve many clients' model objects through a single Engine object.
    """

    def __init__(
        self, tokenizer: Tokenizer, enable_backtrack=True, enable_ff_tokens=True, enable_monitoring=True, **kwargs
    ):
        from ...registry import get_monitor

        self.tokenizer = tokenizer
        self._enable_backtrack = enable_backtrack
        self._enable_ff_tokens = enable_ff_tokens
        self._enable_monitoring = enable_monitoring
        self._top_k = kwargs.get("top_k", 5)

        if self._enable_monitoring:
            # Idempotent start
            _ = get_monitor()

        log_init(f"engine({id(self)})")

    # These need to be properties because once an Engine is started, you can't change their behavior.
    @property
    def enable_backtrack(self):
        return self._enable_backtrack

    @property
    def enable_ff_tokens(self):
        return self._enable_ff_tokens

    @property
    def enable_monitoring(self):
        return self._enable_monitoring

    def get_chat_template(
        self,
    ):  # TODO [HN]: Add more logic here...should we instantiate class here? do we even need to?
        return self.tokenizer.chat_template()  # Instantiate the class before returning to client for now

    def __call__(
        self,
        state: EngineState,
        grammar: str,
        ensure_bos_token: bool = True,
        echo: bool = True,
        sampling_params: Optional[SamplingParams] = None,
    ) -> Generator[EngineResponse, None, TokenUsage]:
        """Main entry point for the inference-parser loop. Yields EngineCallResponse objects as
        the parser advances through the grammar.

        Parameters
        ----------
        prompt : str or Parser
            This is represents the current state of a guidance parser that will be extended
            using the passed grammar. If a string is given then we assume the previous parser
            state is just a fixed string prompt, if a full Parser is given then we extend that
            parser by appending the new grammar to the parser's current grammar and then
            inferencing the model. (TODO: implement full parser extension support)
        grammar: Function
            Grammar (RawFunction or GrammarFunction) used to extend the prompt.
        ensure_bos_token: bool
            Ensures that the prompt ends with the BOS token.
        sampling_params: Optional[SamplingParams]
            Additional sampling parameters to apply to the logits.
        """
        # Note: t0 will get reset further down in the loop, just after the break condition
        _t0 = time.time()
        t0 = _t0

        # TODO: Pass these to get_logits
        # images = state.images
        # audio = state.audio
        # videos = state.videos

        tokens = self.tokenizer.encode(state.prompt.encode("utf-8"))

        parser = TokenParser(
            grammar,
            tokenizer=self.tokenizer,
            enable_backtrack=self.enable_backtrack,
            enable_ff_tokens=self.enable_ff_tokens,
        )

        last_temperature = 1.0
        issued_token: Optional[GenToken] = None
        usage = TokenUsage(round_trips=1, ff_tokens=0)

        while not parser.done():
            recode = False
            if issued_token is None:
                prefix_tokens, backtrack, ff_tokens, mask_fut = parser.process_prompt(
                    prompt_tokens=tokens,
                    ensure_bos_token=ensure_bos_token,
                )
                if prefix_tokens:
                    tokens = prefix_tokens + tokens
                    recode = True
            else:
                backtrack, ff_tokens, mask_fut = parser.advance(token_id=issued_token.token_id)

            if backtrack:
                backtracked_bytes = self.tokenizer.decode(tokens[-backtrack:])
                tokens = tokens[:-backtrack]
            else:
                backtracked_bytes = b""
            tokens += ff_tokens

            if recode:
                # Only necessary when we add a prefix (bos token), which can only happen once
                # per loop. Needs to happen after adding ff_tokens to maintain associativity of
                # (model + prompt) + grammar == model + (prompt + grammar)
                tokens = self.tokenizer.recode(tokens)

            # We can avoid a final get_logits call in the case that:
            # 1. The parser has a pending stop
            # 2. There are no ff_tokens (except for our last generated token)
            # TODO: allow avoiding final forward pass if metrics are disabled
            # and we have a pending stop
            if parser.has_pending_stop() and (
                (not ff_tokens)
                or (len(ff_tokens) == 1 and issued_token is not None and ff_tokens[0] == issued_token.token_id)
            ):
                logits = None
                logits_lat_ms = 0.0
            else:
                t1 = time.time()
                logits_output = self.get_logits(token_ids=tokens, include_all_uncached_tokens=True)
                logits = logits_output["logits"]
                usage.input_tokens += logits_output["n_tokens"]
                usage.cached_input_tokens += logits_output["n_cached"]
                if logits_output["n_tokens"] > logits_output["n_cached"]:
                    usage.forward_passes += 1
                else:
                    usage.cached_output_tokens += 1
                logits_lat_ms = (time.time() - t1) * 1000

            # Important: don't wait on this future until after getting the logits;
            # this allows the mask to be built concurrently with model inference
            mask, ll_response = mask_fut.result()
            legacy_engine_response = ll_response.progress.to_engine_call_response()

            if logits is not None:
                # Not the last one -- that's for the *next* token.
                ff_logits = logits[-len(ff_tokens) - 1 : -1]
                ff_probs = (
                    softmax(ff_logits)
                    if last_temperature < _TEMPERATURE_EPSILON
                    else softmax(ff_logits / last_temperature)
                )

                if len(ff_tokens) == len(tokens) and ff_probs.shape[0] == len(ff_tokens) - 1:
                    # We didn't have a BOS token, so we need to fake the first token prob (say... 1?)
                    ff_probs = np.pad(ff_probs, [(1, 0), (0, 0)], mode="constant", constant_values=1.0)
                elif ff_probs.shape[0] < len(ff_tokens):
                    # Not enough logits were returned despite include_all_uncached_tokens=True, probably due to
                    # using a mock model that doesn't bother to return logits for uncached tokens (all are uncached...)
                    ff_probs = np.pad(
                        ff_probs,
                        [(len(ff_tokens) - ff_probs.shape[0], 0), (0, 0)],
                        mode="constant",
                        constant_values=np.nan,
                    )
            else:
                # really just for mypy -- we shouldn't need this
                ff_probs = np.empty(shape=())

            ff_lat_ms = (time.time() - t0) * 1000
            if not ll_response.stop:
                # If we're not stopping, the logit latency will go into the next generated token
                ff_lat_ms -= logits_lat_ms

            gen_tokens = []
            if issued_token is None:
                # Note: intentionally not counting ff_tokens here, as we're counting these
                # all just as input tokens
                for i, token_id in enumerate(ff_tokens):
                    gen_tokens.append(
                        GenToken(
                            token_id=token_id,
                            bytes=self.tokenizer.decode([token_id]),
                            prob=ff_probs[i, token_id],
                            latency_ms=ff_lat_ms / len(ff_tokens),
                            is_input=True,
                        )
                    )
            else:
                gen_tokens.append(issued_token)
                if backtrack or ff_tokens[:1] != [issued_token.token_id]:
                    issued_token.is_backtracked = True
                    ff_start_index = 0
                else:
                    ff_start_index = 1
                ff_tokens = ff_tokens[ff_start_index:]

                # Just update ff tokens here -- usage for issued_token has already been
                # handled where we got logits above
                usage.ff_tokens += len(ff_tokens)
                for i, token_id in enumerate(ff_tokens, start=ff_start_index):
                    gen_tokens.append(
                        GenToken(
                            token_id=token_id,
                            bytes=self.tokenizer.decode([token_id]),
                            prob=ff_probs[i, token_id],
                            latency_ms=ff_lat_ms / len(ff_tokens),
                            is_force_forwarded=True,
                        )
                    )

            engine_response = EngineResponse(
                new_bytes=legacy_engine_response.new_bytes,
                backtrack_bytes=backtracked_bytes,
                capture_groups=legacy_engine_response.capture_groups,
                capture_group_log_probs=legacy_engine_response.capture_group_log_probs,
                backtrack=backtrack,
                tokens=gen_tokens,
            )

            # process engine_response
            # NOTE (loc): We should not yield the engine_response if new_bytes are invalid utf-8 bytes
            # delayed bytes should be handled here in the engine
            yield engine_response

            if ll_response.stop:
                assert mask is None
                # May raise an exception if the parser is in an bad state!
                parser.cleanup()
                # Ensure we break AFTER yielding the final response
                break
            # Reset time down here instead of at the top of the loop in order to make sure
            # we take sampling time into account
            t0 = time.time()

            # Help the type checker: assert that everything we need to get the next token is not None
            assert logits is not None
            assert mask is not None
            assert ll_response.temperature is not None

            can_finish_early = parser.is_accepting() and self.tokenizer.eos_token_id is not None

            if can_finish_early:
                # Type checker needs some help
                assert self.tokenizer.eos_token_id is not None
                # Should be equivalent to parser.is_accepting()
                assert mask[self.tokenizer.eos_token_id]
                # Whenever we are in an accepting state, we will allow the model to generate whatever it wants
                # but we will treat any "illegal" tokens as EOS, allowing the model to finish gracefully.
                # Hence, mask must be None
                mask_for_sampling = None
            else:
                mask_for_sampling = mask

            issued_token = self.get_next_token_with_top_k(
                logits=logits[-1, :],
                logits_lat_ms=logits_lat_ms,
                token_ids=tokens,
                mask=mask_for_sampling,
                temperature=ll_response.temperature,
                k=self._top_k,
                force_return_unmasked_probs=echo,
                sampling_params=sampling_params,
            )
            last_temperature = ll_response.temperature

            if can_finish_early and not mask[issued_token.token_id]:
                # Type checker needs some help
                assert self.tokenizer.eos_token_id is not None
                issued_token.token_id = self.tokenizer.eos_token_id

        usage.total_latency_ms += (time.time() - _t0) * 1000
        return usage

    def get_next_token_with_top_k(
        self,
        logits: NDArray,
        logits_lat_ms: float,
        token_ids: list[int],
        mask: Optional[bytes],
        temperature: float,
        k: int = 5,
        force_return_unmasked_probs: bool = False,
        sampling_params: Optional[SamplingParams] = None,
    ) -> GenTokenExtra:
        """Get the next token and associated top-k tokens from the engine.

        Parameters
        -------
        logits : Optional[NDArray]
            The logits for the current token ids in the sequence.
            If None, the model will call get_logits to get the logits.
        logits_lat_ms: Optional[float]
            The time taken to compute the logits.
            If logits is None, the model will call get_logits to measure the time.
        token_ids : list[int]
            The current token ids in the sequence.
        mask : Optional[bytes]
            The mask to apply to the logits.
        temperature : float
            The temperature to apply to the logits.
        k : int
            The number of top-k tokens to return.
        force_return_unmasked_probs: bool
            If True, the top-k unmasked probabilities will be returned.
        sampling_params : Optional[SamplingParams]
            Additional sampling parameters to apply to the logits.

        Returns
        -------
        EngineOutput
            The output from the model.
        """

        def get_top_k(_probs: NDArray, _k: int = 5) -> list[int]:
            top_k_indices = _probs.argpartition(-_k)[-_k:]
            # Sort by probability in descending order, as above argpartition
            # does not guarantee order. Sorting the smaller array is faster.
            return sorted(top_k_indices, key=lambda idx: _probs[idx], reverse=True)

        def apply_temp_and_sampling_params(
            _logits: NDArray,
            _sampling_params: Optional[SamplingParams],
        ) -> NDArray:
            """Apply the sampling parameters to the logits."""
            if _sampling_params is None:
                return _logits
            _logits = apply_repetition_penalty(token_ids, _logits, _sampling_params)
            _logits = _logits if temperature < _TEMPERATURE_EPSILON else _logits / temperature
            _logits = apply_min_p_filter(_logits, _sampling_params)
            _logits = apply_top_k_and_top_p_filter(_logits, _sampling_params)
            return _logits

        # TODO: only get unmasked probs if we're either echoing or if we have no mask
        filtered_logits = apply_temp_and_sampling_params(logits, sampling_params)
        probs = softmax(filtered_logits)

        top_k: list[int] = []
        if force_return_unmasked_probs:
            # compute top-k without masking
            top_k = get_top_k(probs, k)

        # compute top-k with masking
        masked_top_k: list[int] = []
        masked_probs: Optional[NDArray] = None
        if mask is not None:
            mask = np.frombuffer(mask, dtype=np.uint8)
            masked_logits = np.where(mask != 0, logits, -np.inf)
            filtered_masked_logits = apply_temp_and_sampling_params(masked_logits, sampling_params)
            masked_probs = softmax(filtered_masked_logits)
            masked_top_k = get_top_k(masked_probs, max(k, 1))  # Ensure at least one token is returned

        if temperature < _TEMPERATURE_EPSILON:
            # Greedy sampling
            if mask is not None:
                assert len(masked_top_k) > 0, "Masked top-k should not be empty when mask is provided"
                issued_token = masked_top_k[0]
            else:
                if len(top_k) == 0:
                    top_k = get_top_k(probs, 1)
                issued_token = top_k[0]
        else:
            # We need to sample from the probabilities
            if mask is None:
                issued_token = np.random.choice(len(probs), p=probs)
            else:
                assert masked_probs is not None, "Masked probabilities should not be None when mask is provided"
                issued_token = np.random.choice(len(masked_probs), p=masked_probs)

        top_k_tokens = [
            GenToken(
                token_id=token_id,
                prob=prob,
                bytes=self.tokenizer.decode([token_id]),
                latency_ms=logits_lat_ms,
                is_generated=True,
                is_masked=mask is not None and bool(mask[token_id] == 0),
            )
            for token_id in masked_top_k + top_k
            # Use unmasked probs always
            if (prob := float(probs[token_id])) > 0
        ]
        top_k_tokens = sorted(top_k_tokens, key=lambda t: t.prob, reverse=True)[:k]

        return GenTokenExtra(
            token_id=issued_token,
            # Use unmasked probs always (TODO: maybe we can avoid it if echo=False?)
            prob=float(probs[issued_token]),
            bytes=self.tokenizer.decode([issued_token]),
            latency_ms=logits_lat_ms,
            is_generated=True,
            top_k=top_k_tokens,
        )

    @abstractmethod
    def get_logits(self, token_ids: list[int], include_all_uncached_tokens: bool = False) -> LogitsOutput:
        """
        Get the logits for the given token ids.
        If include_all_uncached_tokens is True:
            logits for all uncached tokens will be returned, i.e.
            the return value's shape will be `(len(tokens)-num_cached, vocab_size)`.
        If include_all_uncached_tokens is False:
            logits for the last token will be returned, i.e.
            the return value's shape will be `(1, vocab_size)`.
        """
        pass
